from NEMO_osticket.apps import OsTicketConfig


OSTICKET_DB_NAME = "osticket"


class OsTicketDatabaseRouter:

    def db_for_read(self, model, **hints):
        if model._meta.app_label == OsTicketConfig.name:
            return OSTICKET_DB_NAME
        return None

    def db_for_write(self, model, **hints):
        if model._meta.app_label == OsTicketConfig.name:
            return OSTICKET_DB_NAME
        return None

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        if app_label == OsTicketConfig.name:
            return False
        return db == "default"



import re
from logging import getLogger
from typing import Dict, Optional

from NEMO.decorators import customization
from NEMO.exceptions import InvalidCustomizationException
from NEMO.models import Tool, UsageEvent, User
from NEMO.views.customization import CustomizationBase
from django.core.exceptions import ValidationError
from django.core.validators import validate_comma_separated_integer_list
from django.template import Context, Template

from NEMO_osticket.models import OstFormField, OstHelpTopic, OstListItems

osticket_customization_logger = getLogger(__name__)


@customization(key="osticket", title="OsTicket")
class OsTicketCustomization(CustomizationBase):
    variables = {
        "osticket_tool_control_replace_problem_tab": "",
        "osticket_tool_control_tab_name": "Help desk",
        "osticket_create_ticket_include_usage_events": "",
        "osticket_create_ticket_include_projects": "",
        "osticket_create_ticket_help_topic_label": "Help Topic",
        "osticket_create_ticket_help_topic_exclude_list": "",
        "osticket_create_ticket_subject_template": "{% if tool %}[{{ tool.name }}] {% endif %}{{ subject }}",
        "osticket_create_ticket_message_template": "{% if tool %}\nTool name: {{ tool.name }}\n<br>\n{% endif %}\nCreated by: {{ user.get_name }}\n<br>\n{% if usage_event %}\nTool usage: {{ usage_event.start|date:'SHORT_DATETIME_FORMAT' }} to {{ usage_event.end|date:'SHORT_DATETIME_FORMAT' }}\n<br>\n{% endif %}\n{% if project %}\nProject: {{ project.name }}\n<br>\n{% endif %}\nHelp topic: {{ topic }}\n<br><br>\n\n{{ message }}",
        "osticket_create_ticket_internal_note_template": "{% load custom_tags_and_filters %}{% if adjustment_request %}To approve or deny this adjustment request in NEMO, click on <a href='{% absolute_url \"edit_adjustment_request\" adjustment_request.id %}'>the following link</a>.{% endif %}",
        "osticket_tool_matching_field_id": "",
        "osticket_tool_matching_nemo_property_template": "{{ tool.id }}",
        "osticket_tool_matching_property_extract_re": "",
        "osticket_create_ticket_create_adjustment_request_option": "",
        "osticket_create_ticket_create_adjustment_request_template": "{% if tool %}Tool name: {{ tool.name }}\n{% endif %}{% if usage_event %}Tool usage: {{ usage_event.start|date:'SHORT_DATETIME_FORMAT' }} to {{ usage_event.end|date:'SHORT_DATETIME_FORMAT' }}\n{% endif %}{% if project %}Project: {{ project.name }}\n{% endif %}Help topic: {{ topic }}\n\n{{ message }}",
    }

    def context(self) -> Dict:
        from NEMO_osticket.osticket_api import get_osticket_form_fields

        context_dict = super().context()
        context_dict["ticket_form_fields"] = get_osticket_form_fields(exclude_matching=False)
        context_dict["help_topics"] = OstHelpTopic.objects.all()
        context_dict["excluded_help_topic_ids"] = self.get_list_int("osticket_create_ticket_help_topic_exclude_list")
        return context_dict

    def validate(self, name, value):
        if name == "osticket_create_ticket_help_topic_exclude_list" and value:
            validate_comma_separated_integer_list(value)
        if name == "osticket_tool_matching_nemo_property_template" and value:
            try:
                Template(value).render(Context({"tool": Tool.objects.first()}))
            except Exception as e:
                raise ValidationError(str(e))
        if value and name in [
            "osticket_create_ticket_subject_template",
            "osticket_create_ticket_message_template",
            "osticket_create_ticket_create_adjustment_request_template",
        ]:
            try:
                Template(value).render(
                    Context(
                        {
                            "subject": "Subject",
                            "message": "Message",
                            "tool": Tool.objects.first(),
                            "topic": "Topic",
                            "user": User.objects.first(),
                            "usage_event": UsageEvent.objects.first(),
                        }
                    )
                )
            except Exception as e:
                raise ValidationError(str(e))

    def save(self, request, element=None) -> Dict[str, Dict[str, str]]:
        errors = super().save(request, element)
        excluded_topics = ",".join(request.POST.getlist("osticket_create_ticket_help_topic_exclude_list", []))
        try:
            self.validate("osticket_create_ticket_help_topic_exclude_list", excluded_topics)
            type(self).set("osticket_create_ticket_help_topic_exclude_list", excluded_topics)
        except (ValidationError, InvalidCustomizationException) as e:
            errors["osticket_create_ticket_help_topic_exclude_list"] = {
                "error": str(e.message or e.msg),
                "value": excluded_topics,
            }
        return errors

    @classmethod
    def get_osticket_tool_matching_field(cls) -> Optional[OstFormField]:
        return OstFormField.objects.filter(id=cls.get_int("osticket_tool_matching_field_id") or None).first()

    @classmethod
    def get_osticket_tool_matching_value(cls, tool: Tool) -> Optional[OstListItems]:
        field = cls.get_osticket_tool_matching_field()
        osticket_re_matching = cls.get("osticket_tool_matching_property_extract_re")
        if field:
            if field.is_list_type():
                list_options = field.get_list_options()
                if list_options:
                    for item in list_options:
                        if cls.is_tool_match(tool, item.value, osticket_re_matching):
                            return item
        return None

    @classmethod
    def get_tool_matching_nemo_property(cls, tool: Tool) -> str:
        return Template(cls.get("osticket_tool_matching_nemo_property_template")).render(Context({"tool": tool}))

    @classmethod
    def is_tool_match(cls, tool: Tool, osticket_field_value: str, osticket_re_matching=None):
        try:
            nemo_property = cls.get_tool_matching_nemo_property(tool)
            osticket_property = None
            if osticket_re_matching:
                match = re.match(rf"{osticket_re_matching}", osticket_field_value)
                if match:
                    osticket_property = match.group(1)
            if nemo_property and osticket_property:
                return nemo_property == osticket_property
        except Exception as e:
            osticket_customization_logger.error(e)
        return False



from django.apps import AppConfig
from django.conf import settings

# Add our custom database router to redirect to the proper database
if (
    hasattr(settings, "DATABASE_ROUTERS")
    and "NEMO_osticket.routers.OsTicketDatabaseRouter" not in settings.DATABASE_ROUTERS
):
    settings.DATABASE_ROUTERS.append("NEMO_osticket.routers.OsTicketDatabaseRouter")
else:
    setattr(settings, "DATABASE_ROUTERS", ["NEMO_osticket.routers.OsTicketDatabaseRouter"])


class OsTicketConfig(AppConfig):
    name = "NEMO_osticket"
    verbose_name = "osTicket"
    default_auto_field = "django.db.models.AutoField"

    def ready(self):
        from NEMO_osticket.osticket_api import check_and_save_osticket_icon

        check_and_save_osticket_icon()



import json

from NEMO.models import Project, Tool, UsageEvent
from django import forms
from django.conf import settings
from django.core.exceptions import ValidationError

from NEMO_osticket.customizations import OsTicketCustomization
from NEMO_osticket.models import OstHelpTopic
from NEMO_osticket.osticket_api import OsTicketSearchObject, get_osticket_form_fields


# This custom model choice field doesn't use the field queryset to check validity
class CustomModelChoiceField(forms.ModelChoiceField):
    def to_python(self, value):
        if value in self.empty_values:
            return None
        try:
            key = self.to_field_name or "pk"
            if isinstance(value, self.queryset.model):
                value = getattr(value, key)
            value = self.queryset.model.objects.all().get(**{key: value})
        except (ValueError, TypeError, self.queryset.model.DoesNotExist):
            raise ValidationError(
                self.error_messages["invalid_choice"],
                code="invalid_choice",
                params={"value": value},
            )
        return value


class OsTicketForm(forms.Form):
    message = forms.CharField(widget=forms.Textarea())
    usage_event = CustomModelChoiceField(required=False, label="Tool usage", queryset=UsageEvent.objects.none())
    project = CustomModelChoiceField(required=False, label="Project", queryset=Project.objects.none())
    create_adjustment_request = forms.BooleanField(required=False, label="Also create an adjustment request")

    def __init__(self, *args, **kwargs):
        create_ticket_include_usage_events = kwargs.pop(
            "include_usage_events", OsTicketCustomization.get_bool("osticket_create_ticket_include_usage_events")
        )
        create_ticket_include_project = kwargs.pop(
            "include_projects", OsTicketCustomization.get_bool("osticket_create_ticket_include_projects")
        )
        super().__init__(*args, **kwargs)
        field_ordering = ["usage_event", "project", "topicId"]
        if not create_ticket_include_usage_events:
            del self.fields["usage_event"]
        if not OsTicketCustomization.get_bool("osticket_create_ticket_create_adjustment_request_option"):
            del self.fields["create_adjustment_request"]
        if not create_ticket_include_project:
            del self.fields["project"]
        topic_choices = []
        if settings.DATABASES.get("osticket", False):
            topic_choices = [
                (item.topic_id, item.topic)
                for item in OstHelpTopic.objects.exclude(
                    topic_id__in=OsTicketCustomization.get_list_int("osticket_create_ticket_help_topic_exclude_list")
                )
            ]
        self.fields["topicId"] = forms.ChoiceField(choices=topic_choices)
        self.fields["topicId"].label = OsTicketCustomization.get("osticket_create_ticket_help_topic_label")
        self.fields["topicId"].required = True
        for ostField in get_osticket_form_fields():
            if ostField.label not in field_ordering:
                field_ordering.append(ostField.name)
            required = ostField.flags not in [13057, 12289, 769]
            max_length = None
            if ostField.configuration:
                config_json = json.loads(ostField.configuration)
                max_length = config_json.get("length", None) or None
            if ostField.type == "text":
                self.fields[ostField.name] = forms.CharField(max_length=max_length)
                self.fields[ostField.name].label = ostField.label
                self.fields[ostField.name].required = required
            elif ostField.type == "memo":
                self.fields[ostField.name] = forms.CharField(max_length=max_length, widget=forms.Textarea())
                self.fields[ostField.name].label = ostField.label
                self.fields[ostField.name].required = required
            elif ostField.is_list_type():
                choices = [(item.id, item.value) for item in ostField.get_list_options()]
                self.fields[ostField.name] = forms.ChoiceField(choices=choices)
                self.fields[ostField.name].label = ostField.label
                self.fields[ostField.name].required = required
            if "message" == ostField.name:
                self.fields["message"].label = ostField.label
        if "subject" in field_ordering:
            field_ordering.remove("subject")
        if "message" in field_ordering:
            field_ordering.remove("message")
        field_ordering.extend(["subject", "message"])
        self.order_fields(field_ordering)


class OsTicketSearchForm(forms.Form):
    email = forms.EmailField(required=True)
    is_open = forms.NullBooleanField(required=False)
    search = forms.CharField(required=False)
    start = forms.DateTimeField(required=False)
    end = forms.DateTimeField(required=False)
    tool = forms.ModelChoiceField(required=False, queryset=Tool.objects.all())

    def get_search_object(self):
        if self.is_valid():
            return OsTicketSearchObject(**self.cleaned_data)
        return None



import hashlib
import os
import re
from datetime import datetime
from logging import getLogger
from typing import Dict, Iterable, Optional, Set, Tuple
from urllib.parse import urljoin

import requests
from NEMO.models import AdjustmentRequest, Project, Tool, UsageEvent, User
from NEMO.typing import QuerySetType
from NEMO.utilities import render_email_template
from NEMO.views.adjustment_requests import send_request_received_email
from NEMO.views.notifications import create_adjustment_request_notification
from django.conf import settings
from django.db import transaction
from django.db.models import Q
from django.utils import timezone
from django.utils.datastructures import OrderedSet

from NEMO_osticket.customizations import OsTicketCustomization
from NEMO_osticket.exceptions import OsTicketException
from NEMO_osticket.models import (
    OstFormEntry,
    OstFormEntryValues,
    OstFormField,
    OstHelpTopic,
    OstListItems,
    OstObjectType,
    OstStaff,
    OstThread,
    OstThreadEntry,
    OstThreadEntryType,
    OstTicket,
    OstUser,
    OstUserEmail,
)

api_logger = getLogger(__name__)


TICKETS_PATH = "/api/tickets.json"


class OsTicketSearchObject:
    def __init__(
        self,
        email: str = None,
        is_open=None,
        search: str = None,
        start: datetime = None,
        end: datetime = None,
        tool: Tool = None,
    ):
        self.email = email
        self.is_open = is_open
        self.search = search
        self.start = start
        self.end = end
        self.tool = tool
        self.allowed_ids: Optional[Set[int]] = None

    def update_allowed_ids(self, new_allowed_ids: Optional[Iterable[int]]):
        if new_allowed_ids is not None:
            if self.allowed_ids is None:
                self.allowed_ids = set(new_allowed_ids)
            else:
                self.allowed_ids = self.allowed_ids.intersection(set(new_allowed_ids))

    def search_tickets(self, preload=True) -> QuerySetType[OstTicket]:
        if not settings.DATABASES.get("osticket", False):
            return set()
        if self.email:
            self.update_allowed_ids(
                OstTicket.objects.filter(user__ostuseremail__address=self.email).values_list("ticket_id", flat=True)
            )
        # is_open can be True (open), False (closed) or None (both)
        if self.is_open is not None:
            if self.is_open:
                self.update_allowed_ids(
                    OstTicket.objects.filter(status__state="open").values_list("ticket_id", flat=True)
                )
            else:
                self.update_allowed_ids(
                    OstTicket.objects.exclude(status__state="open").values_list("ticket_id", flat=True)
                )
        if self.search:
            threads = OstThreadEntry.objects.filter(
                thread__object_type=OstObjectType.TICKET, body__icontains=self.search
            )
            form_entries = OstFormEntryValues.objects.filter(
                entry__object_type=OstObjectType.TICKET, value__icontains=self.search
            )
            if self.start:
                form_entries = form_entries.filter(entry__created__gte=self.start)
            if self.end:
                form_entries = form_entries.filter(entry__created__lte=self.end)
            if self.allowed_ids is not None:
                threads = threads.filter(thread__object_id__in=self.allowed_ids)
                form_entries = form_entries.filter(entry__object_id__in=self.allowed_ids)
            valid_thread_ticket_ids = list(threads.values_list("thread__object_id", flat=True))
            valid_form_ticket_ids = list(form_entries.values_list("entry__object_id", flat=True))
            self.update_allowed_ids(set(valid_thread_ticket_ids + valid_form_ticket_ids))
        if self.tool:
            self.update_allowed_ids(filter_tickets_for_tool(self.tool, self.allowed_ids))
        if self.allowed_ids is None:
            self.allowed_ids = OstTicket.objects.values_list("ticket_id", flat=True)
        if preload:
            return get_preloaded_os_tickets(self.allowed_ids, self.start, self.end)
        else:
            return OstTicket.objects.filter(ticket_id__in=self.allowed_ids)


def filter_tickets_for_tool(tool: Tool, allowed_ticket_ids=None) -> Set[int]:
    # We are making sure we are only getting tickets linked to the given tool
    from NEMO_osticket.customizations import OsTicketCustomization

    list_item = OsTicketCustomization.get_osticket_tool_matching_value(tool)
    form_field = OsTicketCustomization.get_osticket_tool_matching_field()
    form_entry_values = OstFormEntryValues.objects.filter(field=form_field, entry__object_type=OstObjectType.TICKET)
    if allowed_ticket_ids is not None:
        # Filter by specific ticket ids if available
        form_entry_values = form_entry_values.filter(entry__object_id__in=allowed_ticket_ids)
    ticket_ids = set()
    if list_item:
        # do some pre-filtering even if not perfect to see if the list id is even in there
        # we are checking for both single and double quotes for the id, should narrow it down quite a bit
        for form_entry_value in form_entry_values.filter(
            Q(value__icontains=f"'{list_item.id}'") | Q(value__icontains=f'"{list_item.id}"')
        ).prefetch_related("entry"):
            try:
                if str(list_item.id) == form_entry_value.get_list_item_id():
                    ticket_ids.add(form_entry_value.entry.object_id)
            except Exception as e:
                api_logger.debug(e)
    else:
        # just get corresponding tickets with value
        ticket_ids = (
            form_entry_values.filter(value=OsTicketCustomization.get_tool_matching_nemo_property(tool))
            .values_list("entry__object_id", flat=True)
            .distinct()
        )
    return ticket_ids


def get_preloaded_os_tickets(
    ticket_ids: Iterable[int], start: datetime = None, end: datetime = None, order_by="-created"
) -> Set[OstTicket]:
    """Load ticket with everything we need"""
    tickets: OrderedSet[OstTicket] = OrderedSet()
    threads = list(OstThread.objects.filter(object_id__in=ticket_ids, object_type=OstObjectType.TICKET))
    thread_entries = list(OstThreadEntry.objects.filter(thread__in=[thread.id for thread in threads]))
    form_entries = list(
        OstFormEntry.objects.filter(object_id__in=ticket_ids, object_type=OstObjectType.TICKET).prefetch_related(
            "ostformentryvalues_set__field"
        )
    )
    filtered_tickets = (
        OstTicket.objects.filter(ticket_id__in=ticket_ids).order_by(order_by).prefetch_related("user", "staff", "topic")
    )
    if start:
        filtered_tickets = filtered_tickets.filter(created__gte=start)
    if end:
        filtered_tickets = filtered_tickets.filter(created__lte=end)
    # preload nemo users
    nemo_users = {
        user["username"]: user
        for user in User.objects.filter(
            username__in=[ticket.user.name for ticket in filtered_tickets if ticket.user]
            + [thread_entry.poster for thread_entry in thread_entries]
        ).values("username", "first_name", "last_name")
    }
    for thread_entry in thread_entries:
        if thread_entry.poster:
            thread_entry.nemo_user = nemo_users.get(thread_entry.poster)
    list_items = {str(item.id): item.value for item in OstListItems.objects.all()}
    tool_matching_field_id = OsTicketCustomization.get_int("osticket_tool_matching_field_id")
    for ticket in filtered_tickets:
        # Orderedset is very important here to keep the order
        ticket.ostthread_set = OrderedSet([thread for thread in threads if thread.object_id == ticket.ticket_id])
        ticket.thread = next(iter(ticket.ostthread_set))
        ticket.ostformentry_set = OrderedSet(
            [form_entry for form_entry in form_entries if form_entry.object_id == ticket.ticket_id]
        )
        # Create a map of all entries so we can access them directly later.
        # The map will have the field name, and a tuple of label and value: {"priority": ("Priority Level", "High")}
        for form_entry in form_entries:
            if form_entry.object_id == ticket.ticket_id:
                for form_entry_value in form_entry.ostformentryvalues_set.all():
                    if form_entry_value.field_id != tool_matching_field_id:
                        display_value = form_entry_value.field.label
                        if form_entry_value.field.is_list_type() and form_entry_value.get_list_item_id():
                            value = list_items.get(form_entry_value.get_list_item_id())
                        else:
                            value = form_entry_value.value
                        ticket.form_entry_values[form_entry_value.field.name] = (display_value, value)
        ticket.ostthreadentry_set = OrderedSet(
            [
                thread_entry
                for thread_entry in thread_entries
                if thread_entry.thread_id in [thread.id for thread in ticket.ostthread_set]
            ]
        )
        if ticket.user:
            ticket.nemo_user = nemo_users.get(ticket.user.name)
        ticket.subject()  # preload subject
        tickets.add(ticket)
    return tickets


def create_osticket_user(user: User) -> OstUser:
    """Adds a user to OsTicket"""
    # let's create the user with a fake default email id and assign it later (since it cannot be null and the user cannot be null either on the email object)
    osticket_user = OstUser.objects.create(name=user.username, status=0, default_email_id=0)
    user_email = OstUserEmail.objects.create(user=osticket_user, address=user.email, flags=0)
    osticket_user.default_email_id = user_email.id
    osticket_user.save()
    return osticket_user


def add_reply_to_ticket(request, ticket: OstTicket, user: User, message: str, internal: bool = False):
    """Adds a reply to a ticket"""
    thread = (
        OstThread.objects.filter(object_id=ticket.ticket_id, object_type=OstObjectType.TICKET)
        .prefetch_related("ostthreadentry_set")
        .last()
    )
    username = user.username
    osticket_user, osticket_staff = get_staff_or_user_or_create(user)
    entry_type = OstThreadEntryType.MESSAGE
    # It's only a response if the user is a staff in osticket
    if osticket_staff:
        entry_type = OstThreadEntryType.RESPONSE
    if internal:
        entry_type = OstThreadEntryType.INTERNAL_NOTE
    flags = 64 if internal else 1
    OstThreadEntry.objects.create(
        type=entry_type,
        flags=flags,
        poster=username,
        thread=thread,
        staff=osticket_staff,
        user=osticket_user,
        body=message,
        pid=thread.id,
        format="text",
        ip_address=get_client_ip(request),
    )
    if entry_type == OstThreadEntryType.RESPONSE:
        thread.lastresponse = timezone.now()
        thread.save(update_fields=["lastresponse"])
    elif entry_type == OstThreadEntryType.MESSAGE:
        thread.lastmessage = timezone.now()
        thread.save(update_fields=["lastmessage"])
    ticket.updated = timezone.now()
    ticket.save(update_fields=["updated"])


def get_staff_or_user_or_create(user: User) -> Tuple[OstUser, OstStaff]:
    # This function will try to find a staff in OsTicket ("Agent") or a user, and if none exist will create a user
    osticket_staff = OstStaff.objects.filter(email=user.email).last()
    osticket_user = None
    # if we found the staff, we are done
    if not osticket_staff:
        # otherwise let's find the user
        osticket_user = OstUser.objects.filter(ostuseremail__address=user.email).last()
        # we have a user, we are done
        if not osticket_user:
            # no staff and no user, let's add a user (even if the user is staff in NEMO)
            osticket_user = create_osticket_user(user)
    return osticket_user, osticket_staff


@transaction.atomic
def create_os_ticket(request, data: Dict, tool: Optional[Tool] = None):
    user = request.user
    os_ticket_service = get_os_ticket_service()
    if os_ticket_service.get("available", False):
        try:
            full_tickets_url = os_ticket_service["url"] + TICKETS_PATH
            keyword_arguments = os_ticket_service.get("keyword_arguments", {})
            json_data = data
            # usage_event and project cannot be serialized
            usage_event = json_data.get("usage_event")
            if "usage_event" in json_data:
                del json_data["usage_event"]
            project = json_data.get("project")
            if "project" in json_data:
                del json_data["project"]
            # update with matching form field if applicable
            form_field = OsTicketCustomization.get_osticket_tool_matching_field()
            if form_field:
                if form_field.is_list_type():
                    json_data.update({form_field.name: OsTicketCustomization.get_osticket_tool_matching_value(tool).id})
                else:
                    json_data.update({form_field.name: OsTicketCustomization.get_tool_matching_nemo_property(tool)})
            json_data.update({"email": user.email, "name": user.username})
            topic = OstHelpTopic.objects.filter(topic_id=json_data.get("topicId")).first()
            topic = topic.topic if topic else None

            # Create the corresponding adjustment request in NEMO if the box is checked
            adjustment_request = None
            if (
                OsTicketCustomization.get_bool("osticket_create_ticket_create_adjustment_request_option")
                and "create_adjustment_request" in data
                and data["create_adjustment_request"]
            ):
                data.pop("create_adjustment_request")
                adjustment_request = create_adjustment_request(
                    request, json_data["message"], topic, tool, usage_event, project
                )

            json_data["subject"] = format_ticket_subject(
                request, json_data["subject"], topic, tool, usage_event, project, adjustment_request
            )
            json_data["message"] = "data:text/html," + format_ticket_message(
                request, json_data["message"], topic, tool, usage_event, project, adjustment_request
            )
            response = requests.post(full_tickets_url, json=json_data, **keyword_arguments)
            response.raise_for_status()
            internal_note = OsTicketCustomization.get("osticket_create_ticket_internal_note_template")
            if internal_note.strip():
                message = format_ticket_internal_note(
                    request, json_data["subject"], topic, tool, usage_event, project, adjustment_request
                )
                if message.strip():
                    # We need to add an internal note
                    ticket_number = response.json()
                    add_reply_to_ticket(
                        request, OstTicket.objects.get(number=ticket_number), user, message, internal=True
                    )

        except Exception as e:
            raise OsTicketException(e)
    else:
        raise OsTicketException("OsTicket is not available")


def format_ticket_subject(
    request,
    subject: str,
    topic,
    tool: Tool,
    usage_event: UsageEvent = None,
    project: Project = None,
    adjustment_request: AdjustmentRequest = None,
) -> str:
    user: User = request.user
    template = OsTicketCustomization.get("osticket_create_ticket_subject_template")
    dictionary = {
        "subject": subject,
        "topic": topic,
        "tool": tool,
        "user": user,
        "usage_event": usage_event,
        "project": project,
        "adjustment_request": adjustment_request,
    }
    return render_template(template, dictionary, request)


def format_ticket_message(
    request,
    message: str,
    topic,
    tool: Tool,
    usage_event: UsageEvent = None,
    project: Project = None,
    adjustment_request: AdjustmentRequest = None,
) -> str:
    user: User = request.user
    template = OsTicketCustomization.get("osticket_create_ticket_message_template")
    dictionary = {
        "message": message,
        "topic": topic,
        "tool": tool,
        "user": user,
        "usage_event": usage_event,
        "project": project,
        "adjustment_request": adjustment_request,
    }
    return render_template(template, dictionary, request)


def format_ticket_internal_note(
    request,
    message: str,
    topic,
    tool: Tool,
    usage_event: UsageEvent = None,
    project: Project = None,
    adjustment_request: AdjustmentRequest = None,
) -> str:
    user: User = request.user
    template = OsTicketCustomization.get("osticket_create_ticket_internal_note_template")
    dictionary = {
        "message": message,
        "topic": topic,
        "tool": tool,
        "user": user,
        "usage_event": usage_event,
        "project": project,
        "adjustment_request": adjustment_request,
    }
    return render_template(template, dictionary, request)


def create_adjustment_request(
    request, message: str, topic, tool: Tool = None, usage_event: UsageEvent = None, project: Project = None
) -> AdjustmentRequest:
    user: User = request.user
    template = OsTicketCustomization.get("osticket_create_ticket_create_adjustment_request_template")
    dictionary = {
        "message": message,
        "topic": topic,
        "tool": tool,
        "user": user,
        "usage_event": usage_event,
        "project": project,
    }
    adjustment_request_description = render_template(template, dictionary, request)
    adjustment_request = AdjustmentRequest.objects.create(
        creator=user, description=adjustment_request_description, item=usage_event
    )
    create_adjustment_request_notification(adjustment_request)
    reviewers: Set[User] = set(list(adjustment_request.reviewers()))
    send_request_received_email(request, adjustment_request, False, reviewers)
    return adjustment_request


def render_template(template, dictionary, request) -> str:
    return render_email_template(template, dictionary, request)


def get_client_ip(request):
    x_forwarded_for = request.META.get("HTTP_X_FORWARDED_FOR")
    if x_forwarded_for:
        ip = x_forwarded_for.split(",")[0]
    else:
        ip = request.META.get("REMOTE_ADDR")
    return ip


def get_osticket_form_fields(exclude_matching=True) -> Set[OstFormField]:
    # we need to exclude the tool matching field since we will be setting it automatically
    form_fields = set()
    if settings.DATABASES.get("osticket", False):
        form_fields = OstFormField.objects.filter(form__type=OstObjectType.TICKET)
        if exclude_matching:
            form_fields = form_fields.exclude(
                id=OsTicketCustomization.get_int("osticket_tool_matching_field_id") or None
            )
    return form_fields


def get_ticket_form_field_labels() -> Dict[str, str]:
    return {field.name: field.label for field in get_osticket_form_fields()}


def check_and_save_osticket_icon():
    url = get_os_ticket_service().get("url")
    name = "osticket_icon.png"

    base_dir = settings.STATIC_ROOT

    # 1 Try to find the favicon URL from the HTML <link> tag
    try:
        html = requests.get(url, timeout=5).text
        match = re.search(r'<link[^>]+rel=["\'](?:shortcut icon|icon)["\'][^>]*>', html, re.I)
        if match:
            href_match = re.search(r'href=["\'](.*?)["\']', match.group(0), re.I)
            favicon_url = urljoin(url, href_match.group(1)) if href_match else None
        else:
            favicon_url = None
    except:
        favicon_url = None

    # 2 Fallback to /favicon.ico if not found
    if not favicon_url:
        favicon_url = urljoin(url, "/favicon.ico")

    # 3 Download the favicon
    try:
        resp = requests.get(favicon_url, timeout=5)
        if resp.status_code != 200 or not resp.content:
            return
        content = resp.content
    except:
        return

    # 4 Compare with the existing file
    save_path = os.path.join(base_dir, f"{name}")
    new_hash = hashlib.md5(content).hexdigest()

    if os.path.exists(save_path):
        with open(save_path, "rb") as f:
            existing_hash = hashlib.md5(f.read()).hexdigest()
        if existing_hash == new_hash:
            return

    # 5 Save the new file
    with open(save_path, "wb") as f:
        f.write(content)


def get_os_ticket_service():
    return getattr(settings, "OSTICKET_SERVICE", {})



from __future__ import annotations

import json
from typing import Optional

from NEMO.models import User
from NEMO.typing import QuerySetType
from django.db import models


class OstObjectType(object):
    TICKET = "T"
    TICKET_STATUS_PROPERTIES = "L1"
    TASK = "A"
    CUSTOM = "G"
    COMPANY = "C"
    ORGANIZATION = "C"
    CONTACT = "U"


class OstThreadEntryType(object):
    MESSAGE = "M"
    RESPONSE = "R"
    INTERNAL_NOTE = "N"


class MySQLZeroRelatedFieldMixin:
    """
    ForeignKey for legacy MYSQL databases that use 0 instead of NULL.
    - Converts None → 0 when saving
    - Converts 0 → None when reading
    """

    def __init__(self, *args, **kwargs):
        # These are always logically "nullable" at the ORM level
        kwargs.setdefault("null", True)
        kwargs.setdefault("blank", True)
        super().__init__(*args, **kwargs)

    def get_prep_value(self, value):
        """
        Called before saving to DB.
        Converts None → 0 so the DB gets '0' instead of NULL.
        """
        if value is None:
            return 0
        return super().get_prep_value(value)

    def get_db_prep_save(self, value, connection):
        # Ensures DB gets 0, not NULL, even if None slipped through
        # This also temporarily enables 0 for auto pk
        # due to the django validation checks on the related model pk
        if value is None:
            value = 0
        old_value = connection.features.allows_auto_pk_0
        connection.features.allows_auto_pk_0 = True
        prepped_value = super().get_db_prep_save(value, connection)
        connection.features.allows_auto_pk_0 = old_value
        return prepped_value

    def from_db_value(self, value, expression, connection):
        """
        Called when reading from DB.
        Converts 0 → None for ORM usage.
        """
        if value == 0:
            return None
        return value


class MySQLZeroForeignKey(MySQLZeroRelatedFieldMixin, models.ForeignKey):
    pass


class MySQLZeroOneToOne(MySQLZeroRelatedFieldMixin, models.OneToOneField):
    pass


class OstSearch(models.Model):
    object_type = models.CharField(
        primary_key=True, max_length=8
    )  # The composite primary key (object_type, object_id) found, that is not supported. The first column is selected.
    object_id = models.PositiveIntegerField()
    title = models.TextField(blank=True, null=True)
    content = models.TextField(blank=True, null=True)

    class Meta:
        managed = False
        db_table = "ost__search"
        unique_together = (("object_type", "object_id"),)


class OstApiKey(models.Model):
    isactive = models.IntegerField()
    ipaddr = models.CharField(max_length=64)
    apikey = models.CharField(unique=True, max_length=255)
    can_create_tickets = models.PositiveIntegerField()
    can_exec_cron = models.PositiveIntegerField()
    notes = models.TextField(blank=True, null=True)
    updated = models.DateTimeField()
    created = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_api_key"


class OstAttachment(models.Model):
    object_id = models.PositiveIntegerField()
    type = models.CharField(max_length=1)
    file_id = models.PositiveIntegerField()
    name = models.CharField(max_length=255, blank=True, null=True)
    inline = models.PositiveIntegerField()
    lang = models.CharField(max_length=16, blank=True, null=True)

    class Meta:
        managed = False
        db_table = "ost_attachment"
        unique_together = (
            ("file_id", "object_id"),
            ("object_id", "file_id", "type"),
        )


class OstCannedResponse(models.Model):
    canned_id = models.AutoField(primary_key=True)
    # dept_id = models.PositiveIntegerField()
    dept = models.ForeignKey("OstDepartment", null=True, blank=True, on_delete=models.DO_NOTHING)
    isenabled = models.PositiveIntegerField()
    title = models.CharField(unique=True, max_length=255)
    response = models.TextField()
    lang = models.CharField(max_length=16)
    notes = models.TextField(blank=True, null=True)
    created = models.DateTimeField()
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_canned_response"


class OstConfig(models.Model):
    namespace = models.CharField(max_length=64)
    key = models.CharField(max_length=64)
    value = models.TextField()
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_config"
        unique_together = (("namespace", "key"),)


class OstContent(models.Model):
    isactive = models.PositiveIntegerField()
    type = models.CharField(max_length=32)
    name = models.CharField(unique=True, max_length=255)
    body = models.TextField()
    notes = models.TextField(blank=True, null=True)
    created = models.DateTimeField()
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_content"


class OstDepartment(models.Model):
    pid = models.PositiveIntegerField(blank=True, null=True)
    # tpl_id = models.PositiveIntegerField()
    tpl = models.ForeignKey("OstEmailTemplateGroup", null=True, blank=True, on_delete=models.DO_NOTHING)
    # sla_id = models.PositiveIntegerField()
    sla = models.ForeignKey("OstSla", null=True, blank=True, on_delete=models.DO_NOTHING)
    # schedule_id = models.PositiveIntegerField()
    schedule = models.ForeignKey("OstSchedule", null=True, blank=True, on_delete=models.DO_NOTHING)
    email_id = models.PositiveIntegerField()
    autoresp_email_id = models.PositiveIntegerField()
    manager_id = models.PositiveIntegerField()
    flags = models.PositiveIntegerField()
    name = models.CharField(max_length=128)
    signature = models.TextField()
    ispublic = models.PositiveIntegerField()
    group_membership = models.IntegerField()
    ticket_auto_response = models.IntegerField()
    message_auto_response = models.IntegerField()
    path = models.CharField(max_length=128)
    updated = models.DateTimeField()
    created = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_department"
        unique_together = (("name", "pid"),)


class OstDraft(models.Model):
    staff_id = models.PositiveIntegerField()
    namespace = models.CharField(max_length=32)
    body = models.TextField()
    extra = models.TextField(blank=True, null=True)
    created = models.DateTimeField()
    updated = models.DateTimeField(blank=True, null=True)

    class Meta:
        managed = False
        db_table = "ost_draft"


class OstEmail(models.Model):
    email_id = models.AutoField(primary_key=True)
    noautoresp = models.PositiveIntegerField()
    # priority_id = models.PositiveIntegerField()
    priority = models.ForeignKey("OstTicketPriority", null=True, blank=True, on_delete=models.DO_NOTHING)
    # dept_id = models.PositiveIntegerField()
    dept = models.ForeignKey(OstDepartment, null=True, blank=True, on_delete=models.DO_NOTHING)
    # topic_id = models.PositiveIntegerField()
    topic = models.ForeignKey("OstHelpTopic", null=True, blank=True, on_delete=models.DO_NOTHING)
    email = models.CharField(unique=True, max_length=255)
    name = models.CharField(max_length=255)
    userid = models.CharField(max_length=255)
    userpass = models.CharField(max_length=255, db_collation="ascii_general_ci")
    mail_active = models.IntegerField()
    mail_host = models.CharField(max_length=255)
    mail_protocol = models.CharField(max_length=4)
    mail_encryption = models.CharField(max_length=4)
    mail_folder = models.CharField(max_length=255, blank=True, null=True)
    mail_port = models.IntegerField(blank=True, null=True)
    mail_fetchfreq = models.IntegerField()
    mail_fetchmax = models.IntegerField()
    mail_archivefolder = models.CharField(max_length=255, blank=True, null=True)
    mail_delete = models.IntegerField()
    mail_errors = models.IntegerField()
    mail_lasterror = models.DateTimeField(blank=True, null=True)
    mail_lastfetch = models.DateTimeField(blank=True, null=True)
    smtp_active = models.IntegerField(blank=True, null=True)
    smtp_host = models.CharField(max_length=255)
    smtp_port = models.IntegerField(blank=True, null=True)
    smtp_secure = models.IntegerField()
    smtp_auth = models.IntegerField()
    smtp_auth_creds = models.IntegerField(blank=True, null=True)
    smtp_userid = models.CharField(max_length=255)
    smtp_userpass = models.CharField(max_length=255, db_collation="ascii_general_ci")
    smtp_spoofing = models.PositiveIntegerField()
    notes = models.TextField(blank=True, null=True)
    created = models.DateTimeField()
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_email"


class OstEmailAccount(models.Model):
    name = models.CharField(max_length=128)
    active = models.IntegerField()
    protocol = models.CharField(max_length=64)
    host = models.CharField(max_length=128)
    port = models.IntegerField()
    username = models.CharField(max_length=128, blank=True, null=True)
    password = models.CharField(max_length=255, blank=True, null=True)
    options = models.CharField(max_length=512, blank=True, null=True)
    errors = models.PositiveIntegerField(blank=True, null=True)
    created = models.DateTimeField()
    updated = models.DateTimeField()
    lastconnect = models.DateTimeField(blank=True, null=True)
    lasterror = models.DateTimeField(blank=True, null=True)

    class Meta:
        managed = False
        db_table = "ost_email_account"


class OstEmailTemplate(models.Model):
    # tpl_id = models.PositiveIntegerField()
    tpl = models.ForeignKey("OstEmailTemplateGroup", null=True, blank=True, on_delete=models.DO_NOTHING)
    code_name = models.CharField(max_length=32)
    subject = models.CharField(max_length=255)
    body = models.TextField()
    notes = models.TextField(blank=True, null=True)
    created = models.DateTimeField()
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_email_template"
        unique_together = (("tpl_id", "code_name"),)


class OstEmailTemplateGroup(models.Model):
    tpl_id = models.AutoField(primary_key=True)
    isactive = models.PositiveIntegerField()
    name = models.CharField(max_length=32)
    lang = models.CharField(max_length=16)
    notes = models.TextField(blank=True, null=True)
    created = models.DateTimeField()
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_email_template_group"


class OstEvent(models.Model):
    name = models.CharField(unique=True, max_length=60)
    description = models.CharField(max_length=60, blank=True, null=True)

    class Meta:
        managed = False
        db_table = "ost_event"


class OstFaq(models.Model):
    faq_id = models.AutoField(primary_key=True)
    category_id = models.PositiveIntegerField()
    ispublished = models.PositiveIntegerField()
    question = models.CharField(unique=True, max_length=255)
    answer = models.TextField()
    keywords = models.TextField(blank=True, null=True)
    notes = models.TextField(blank=True, null=True)
    created = models.DateTimeField()
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_faq"


class OstFaqCategory(models.Model):
    category_id = models.AutoField(primary_key=True)
    category_pid = models.PositiveIntegerField(blank=True, null=True)
    ispublic = models.PositiveIntegerField()
    name = models.CharField(max_length=125, blank=True, null=True)
    description = models.TextField()
    notes = models.TextField()
    created = models.DateTimeField()
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_faq_category"


class OstFaqTopic(models.Model):
    faq_id = models.PositiveIntegerField(
        primary_key=True
    )  # The composite primary key (faq_id, topic_id) found, that is not supported. The first column is selected.
    topic_id = models.PositiveIntegerField()

    class Meta:
        managed = False
        db_table = "ost_faq_topic"
        unique_together = (("faq_id", "topic_id"),)


class OstFile(models.Model):
    ft = models.CharField(max_length=1)
    bk = models.CharField(max_length=1)
    type = models.CharField(max_length=255, db_collation="ascii_general_ci")
    size = models.PositiveBigIntegerField()
    key = models.CharField(max_length=86, db_collation="ascii_general_ci")
    signature = models.CharField(max_length=86, db_collation="ascii_bin")
    name = models.CharField(max_length=255)
    attrs = models.CharField(max_length=255, blank=True, null=True)
    created = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_file"


class OstFileChunk(models.Model):
    file_id = models.IntegerField(
        primary_key=True
    )  # The composite primary key (file_id, chunk_id) found, that is not supported. The first column is selected.
    chunk_id = models.IntegerField()
    filedata = models.TextField()

    class Meta:
        managed = False
        db_table = "ost_file_chunk"
        unique_together = (("file_id", "chunk_id"),)


class OstFilter(models.Model):
    execorder = models.PositiveIntegerField()
    isactive = models.PositiveIntegerField()
    flags = models.PositiveIntegerField(blank=True, null=True)
    status = models.PositiveIntegerField()
    match_all_rules = models.PositiveIntegerField()
    stop_onmatch = models.PositiveIntegerField()
    target = models.CharField(max_length=5)
    email_id = models.PositiveIntegerField()
    name = models.CharField(max_length=32)
    notes = models.TextField(blank=True, null=True)
    created = models.DateTimeField()
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_filter"


class OstFilterAction(models.Model):
    filter_id = models.PositiveIntegerField()
    sort = models.PositiveIntegerField()
    type = models.CharField(max_length=24)
    configuration = models.TextField(blank=True, null=True)
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_filter_action"


class OstFilterRule(models.Model):
    filter_id = models.PositiveIntegerField()
    what = models.CharField(max_length=32)
    how = models.CharField(max_length=10)
    val = models.CharField(max_length=255)
    isactive = models.PositiveIntegerField()
    notes = models.TextField()
    created = models.DateTimeField()
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_filter_rule"
        unique_together = (("filter_id", "what", "how", "val"),)


class OstForm(models.Model):
    pid = models.PositiveIntegerField(blank=True, null=True)
    type = models.CharField(max_length=8)
    flags = models.PositiveIntegerField()
    title = models.CharField(max_length=255)
    instructions = models.CharField(max_length=512, blank=True, null=True)
    name = models.CharField(max_length=64)
    notes = models.TextField(blank=True, null=True)
    created = models.DateTimeField()
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_form"


class OstFormEntry(models.Model):
    # form_id = models.PositiveIntegerField()
    form = models.ForeignKey(OstForm, on_delete=models.DO_NOTHING)
    object_id = models.PositiveIntegerField(blank=True, null=True)
    object_type = models.CharField(max_length=1)
    sort = models.PositiveIntegerField()
    extra = models.TextField(blank=True, null=True)
    created = models.DateTimeField()
    updated = models.DateTimeField()

    @property
    def ticket(self):
        if self.object_type == OstObjectType.TICKET:
            return OstTicket.objects.get(object_id=self.object_id)
        return None

    class Meta:
        managed = False
        db_table = "ost_form_entry"


class OstFormEntryValues(models.Model):
    # The composite primary key (entry_id, field_id) found, that is not supported. The first column is selected.
    entry_id = models.PositiveIntegerField(primary_key=True)
    entry = models.ForeignKey(OstFormEntry, on_delete=models.DO_NOTHING)
    # field_id = models.PositiveIntegerField()
    field = models.ForeignKey("OstFormField", on_delete=models.DO_NOTHING)
    value = models.TextField(blank=True, null=True)
    value_id = models.IntegerField(blank=True, null=True)

    # We need to override those methods since we have a necessary clash with entry_id
    @classmethod
    def _check_field_name_clashes(cls):
        return []

    @classmethod
    def _check_column_name_clashes(cls):
        return []

    def get_list_item_id(self) -> Optional[int]:
        try:
            return next(iter(json.loads(self.value)))
        except Exception:
            pass
        return None

    class Meta:
        managed = False
        db_table = "ost_form_entry_values"
        unique_together = (("entry_id", "field_id"),)


class OstFormField(models.Model):
    # form_id = models.PositiveIntegerField()
    form = models.ForeignKey(OstForm, on_delete=models.DO_NOTHING)
    flags = models.PositiveIntegerField(blank=True, null=True)
    type = models.CharField(max_length=255)
    label = models.CharField(max_length=255)
    name = models.CharField(max_length=64)
    configuration = models.TextField(blank=True, null=True)
    sort = models.PositiveIntegerField()
    hint = models.CharField(max_length=512, blank=True, null=True)
    created = models.DateTimeField()
    updated = models.DateTimeField()

    def get_list_options(self) -> QuerySetType[OstListItems]:
        if self.is_list_type():
            list_id = self.type.replace("list-", "")
            return OstListItems.objects.filter(list_id=list_id)
        return OstListItems.objects.none()

    def is_list_type(self) -> bool:
        return self.type.startswith("list-")

    class Meta:
        managed = False
        db_table = "ost_form_field"
        ordering = ["sort"]


class OstGroup(models.Model):
    role_id = models.PositiveIntegerField()
    flags = models.PositiveIntegerField()
    name = models.CharField(max_length=120)
    notes = models.TextField(blank=True, null=True)
    created = models.DateTimeField()
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_group"


class OstHelpTopic(models.Model):
    topic_id = models.AutoField(primary_key=True)
    topic_pid = models.PositiveIntegerField()
    ispublic = models.PositiveIntegerField()
    noautoresp = models.PositiveIntegerField()
    flags = models.PositiveIntegerField(blank=True, null=True)
    status_id = models.PositiveIntegerField()
    priority_id = models.PositiveIntegerField()
    dept_id = models.PositiveIntegerField()
    staff_id = models.PositiveIntegerField()
    team_id = models.PositiveIntegerField()
    # sla_id = models.PositiveIntegerField()
    sla = models.ForeignKey("OstSla", null=True, blank=True, on_delete=models.DO_NOTHING)
    page_id = models.PositiveIntegerField()
    sequence_id = models.PositiveIntegerField()
    sort = models.PositiveIntegerField()
    topic = models.CharField(max_length=32)
    number_format = models.CharField(max_length=32, blank=True, null=True)
    notes = models.TextField(blank=True, null=True)
    created = models.DateTimeField()
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_help_topic"
        unique_together = (("topic", "topic_pid"),)


class OstHelpTopicForm(models.Model):
    # topic_id = models.PositiveIntegerField()
    topic = models.ForeignKey(OstHelpTopic, on_delete=models.DO_NOTHING)
    # form_id = models.PositiveIntegerField()
    form = models.ForeignKey(OstForm, on_delete=models.DO_NOTHING)
    sort = models.PositiveIntegerField()
    extra = models.TextField(blank=True, null=True)

    class Meta:
        managed = False
        db_table = "ost_help_topic_form"


class OstList(models.Model):
    name = models.CharField(max_length=255)
    name_plural = models.CharField(max_length=255, blank=True, null=True)
    sort_mode = models.CharField(max_length=7)
    masks = models.PositiveIntegerField()
    type = models.CharField(max_length=16, blank=True, null=True)
    configuration = models.TextField()
    notes = models.TextField(blank=True, null=True)
    created = models.DateTimeField()
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_list"


class OstListItems(models.Model):
    # list_id = models.IntegerField(blank=True, null=True)
    list = models.ForeignKey(OstList, null=True, blank=True, on_delete=models.DO_NOTHING)
    status = models.PositiveIntegerField()
    value = models.CharField(max_length=255)
    extra = models.CharField(max_length=255, blank=True, null=True)
    sort = models.IntegerField()
    properties = models.TextField(blank=True, null=True)

    class Meta:
        managed = False
        db_table = "ost_list_items"


class OstLock(models.Model):
    lock_id = models.AutoField(primary_key=True)
    staff_id = models.PositiveIntegerField()
    expire = models.DateTimeField(blank=True, null=True)
    code = models.CharField(max_length=20, blank=True, null=True)
    created = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_lock"


class OstNote(models.Model):
    pid = models.PositiveIntegerField(blank=True, null=True)
    staff_id = models.PositiveIntegerField()
    ext_id = models.CharField(max_length=10, blank=True, null=True)
    body = models.TextField(blank=True, null=True)
    status = models.PositiveIntegerField()
    sort = models.PositiveIntegerField()
    created = models.DateTimeField()
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_note"


class OstOrganization(models.Model):
    name = models.CharField(max_length=128)
    manager = models.CharField(max_length=16)
    status = models.PositiveIntegerField()
    domain = models.CharField(max_length=256)
    extra = models.TextField(blank=True, null=True)
    created = models.DateTimeField(blank=True, null=True)
    updated = models.DateTimeField(blank=True, null=True)

    class Meta:
        managed = False
        db_table = "ost_organization"


class OstPlugin(models.Model):
    name = models.CharField(max_length=30)
    install_path = models.CharField(max_length=60)
    isphar = models.IntegerField()
    isactive = models.IntegerField()
    version = models.CharField(max_length=64, blank=True, null=True)
    installed = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_plugin"


class OstQueue(models.Model):
    parent_id = models.PositiveIntegerField()
    columns_id = models.PositiveIntegerField(blank=True, null=True)
    sort_id = models.PositiveIntegerField(blank=True, null=True)
    flags = models.PositiveIntegerField()
    staff_id = models.PositiveIntegerField()
    sort = models.PositiveIntegerField()
    title = models.CharField(max_length=60, blank=True, null=True)
    config = models.TextField(blank=True, null=True)
    filter = models.CharField(max_length=64, blank=True, null=True)
    root = models.CharField(max_length=32, blank=True, null=True)
    path = models.CharField(max_length=80)
    created = models.DateTimeField()
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_queue"


class OstQueueColumn(models.Model):
    flags = models.PositiveIntegerField()
    name = models.CharField(max_length=64)
    primary = models.CharField(max_length=64)
    secondary = models.CharField(max_length=64, blank=True, null=True)
    filter = models.CharField(max_length=32, blank=True, null=True)
    truncate = models.CharField(max_length=16, blank=True, null=True)
    annotations = models.TextField(blank=True, null=True)
    conditions = models.TextField(blank=True, null=True)
    extra = models.TextField(blank=True, null=True)

    class Meta:
        managed = False
        db_table = "ost_queue_column"


class OstQueueColumns(models.Model):
    queue_id = models.PositiveIntegerField(
        primary_key=True
    )  # The composite primary key (queue_id, column_id, staff_id) found, that is not supported. The first column is selected.
    column_id = models.PositiveIntegerField()
    staff_id = models.PositiveIntegerField()
    bits = models.PositiveIntegerField()
    sort = models.PositiveIntegerField()
    heading = models.CharField(max_length=64, blank=True, null=True)
    width = models.PositiveIntegerField()

    class Meta:
        managed = False
        db_table = "ost_queue_columns"
        unique_together = (("queue_id", "column_id", "staff_id"),)


class OstQueueConfig(models.Model):
    queue_id = models.PositiveIntegerField(
        primary_key=True
    )  # The composite primary key (queue_id, staff_id) found, that is not supported. The first column is selected.
    staff_id = models.PositiveIntegerField()
    setting = models.TextField(blank=True, null=True)
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_queue_config"
        unique_together = (("queue_id", "staff_id"),)


class OstQueueExport(models.Model):
    queue_id = models.PositiveIntegerField()
    path = models.CharField(max_length=64)
    heading = models.CharField(max_length=64, blank=True, null=True)
    sort = models.PositiveIntegerField()

    class Meta:
        managed = False
        db_table = "ost_queue_export"


class OstQueueSort(models.Model):
    root = models.CharField(max_length=32, blank=True, null=True)
    name = models.CharField(max_length=64)
    columns = models.TextField(blank=True, null=True)
    updated = models.DateTimeField(blank=True, null=True)

    class Meta:
        managed = False
        db_table = "ost_queue_sort"


class OstQueueSorts(models.Model):
    queue_id = models.PositiveIntegerField(
        primary_key=True
    )  # The composite primary key (queue_id, sort_id) found, that is not supported. The first column is selected.
    sort_id = models.PositiveIntegerField()
    bits = models.PositiveIntegerField()
    sort = models.PositiveIntegerField()

    class Meta:
        managed = False
        db_table = "ost_queue_sorts"
        unique_together = (("queue_id", "sort_id"),)


class OstRole(models.Model):
    flags = models.PositiveIntegerField()
    name = models.CharField(unique=True, max_length=64, blank=True, null=True)
    permissions = models.TextField(blank=True, null=True)
    notes = models.TextField(blank=True, null=True)
    created = models.DateTimeField()
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_role"


class OstSchedule(models.Model):
    flags = models.PositiveIntegerField()
    name = models.CharField(max_length=255)
    timezone = models.CharField(max_length=64, blank=True, null=True)
    description = models.CharField(max_length=255)
    created = models.DateTimeField()
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_schedule"


class OstScheduleEntry(models.Model):
    # schedule_id = models.PositiveIntegerField()
    schedule = models.ForeignKey(OstSchedule, on_delete=models.DO_NOTHING)
    flags = models.PositiveIntegerField()
    sort = models.PositiveIntegerField()
    name = models.CharField(max_length=255)
    repeats = models.CharField(max_length=16)
    starts_on = models.DateField(blank=True, null=True)
    starts_at = models.TimeField(blank=True, null=True)
    ends_on = models.DateField(blank=True, null=True)
    ends_at = models.TimeField(blank=True, null=True)
    stops_on = models.DateTimeField(blank=True, null=True)
    day = models.IntegerField(blank=True, null=True)
    week = models.IntegerField(blank=True, null=True)
    month = models.IntegerField(blank=True, null=True)
    created = models.DateTimeField()
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_schedule_entry"


class OstSequence(models.Model):
    name = models.CharField(max_length=64, blank=True, null=True)
    flags = models.PositiveIntegerField(blank=True, null=True)
    next = models.PositiveBigIntegerField()
    increment = models.IntegerField(blank=True, null=True)
    padding = models.CharField(max_length=1, blank=True, null=True)
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_sequence"


class OstSession(models.Model):
    session_id = models.CharField(primary_key=True, max_length=255, db_collation="ascii_general_ci")
    session_data = models.TextField(blank=True, null=True)
    session_expire = models.DateTimeField(blank=True, null=True)
    session_updated = models.DateTimeField(blank=True, null=True)
    user_id = models.CharField(max_length=16, db_comment="osTicket staff/client ID")
    user_ip = models.CharField(max_length=64)
    user_agent = models.CharField(max_length=255)

    class Meta:
        managed = False
        db_table = "ost_session"


class OstSla(models.Model):
    # schedule_id = models.PositiveIntegerField()
    schedule = models.ForeignKey(OstSchedule, null=True, blank=True, on_delete=models.DO_NOTHING)
    flags = models.PositiveIntegerField()
    grace_period = models.PositiveIntegerField()
    name = models.CharField(unique=True, max_length=64)
    notes = models.TextField(blank=True, null=True)
    created = models.DateTimeField()
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_sla"


class OstStaff(models.Model):
    staff_id = models.AutoField(primary_key=True)
    dept_id = models.PositiveIntegerField()
    role_id = models.PositiveIntegerField()
    username = models.CharField(unique=True, max_length=32)
    firstname = models.CharField(max_length=32, blank=True, null=True)
    lastname = models.CharField(max_length=32, blank=True, null=True)
    passwd = models.CharField(max_length=128, blank=True, null=True)
    backend = models.CharField(max_length=32, blank=True, null=True)
    email = models.CharField(max_length=255, blank=True, null=True)
    phone = models.CharField(max_length=24)
    phone_ext = models.CharField(max_length=6, blank=True, null=True)
    mobile = models.CharField(max_length=24)
    signature = models.TextField()
    lang = models.CharField(max_length=16, blank=True, null=True)
    timezone = models.CharField(max_length=64, blank=True, null=True)
    locale = models.CharField(max_length=16, blank=True, null=True)
    notes = models.TextField(blank=True, null=True)
    isactive = models.IntegerField()
    isadmin = models.IntegerField()
    isvisible = models.PositiveIntegerField()
    onvacation = models.PositiveIntegerField()
    assigned_only = models.PositiveIntegerField()
    show_assigned_tickets = models.PositiveIntegerField()
    change_passwd = models.PositiveIntegerField()
    max_page_size = models.PositiveIntegerField()
    auto_refresh_rate = models.PositiveIntegerField()
    default_signature_type = models.CharField(max_length=4)
    default_paper_size = models.CharField(max_length=6)
    extra = models.TextField(blank=True, null=True)
    permissions = models.TextField(blank=True, null=True)
    created = models.DateTimeField()
    lastlogin = models.DateTimeField(blank=True, null=True)
    passwdreset = models.DateTimeField(blank=True, null=True)
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_staff"


class OstStaffDeptAccess(models.Model):
    staff_id = models.PositiveIntegerField(
        primary_key=True
    )  # The composite primary key (staff_id, dept_id) found, that is not supported. The first column is selected.
    dept_id = models.PositiveIntegerField()
    role_id = models.PositiveIntegerField()
    flags = models.PositiveIntegerField()

    class Meta:
        managed = False
        db_table = "ost_staff_dept_access"
        unique_together = (("staff_id", "dept_id"),)


class OstSyslog(models.Model):
    log_id = models.AutoField(primary_key=True)
    log_type = models.CharField(max_length=7)
    title = models.CharField(max_length=255)
    log = models.TextField()
    logger = models.CharField(max_length=64)
    ip_address = models.CharField(max_length=64)
    created = models.DateTimeField()
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_syslog"


class OstTask(models.Model):
    object_id = models.IntegerField()
    object_type = models.CharField(max_length=1)
    number = models.CharField(max_length=20, blank=True, null=True)
    dept_id = models.PositiveIntegerField()
    staff_id = models.PositiveIntegerField()
    team_id = models.PositiveIntegerField()
    lock_id = models.PositiveIntegerField()
    flags = models.PositiveIntegerField()
    duedate = models.DateTimeField(blank=True, null=True)
    closed = models.DateTimeField(blank=True, null=True)
    created = models.DateTimeField()
    updated = models.DateTimeField()

    @property
    def ostthread_set(self):
        return OstThread.objects.filter(object_id=self.id, object_type=OstObjectType.TICKET)

    class Meta:
        managed = False
        db_table = "ost_task"


class OstTeam(models.Model):
    team_id = models.AutoField(primary_key=True)
    lead_id = models.PositiveIntegerField()
    flags = models.PositiveIntegerField()
    name = models.CharField(unique=True, max_length=125)
    notes = models.TextField(blank=True, null=True)
    created = models.DateTimeField()
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_team"


class OstTeamMember(models.Model):
    team_id = models.PositiveIntegerField(
        primary_key=True
    )  # The composite primary key (team_id, staff_id) found, that is not supported. The first column is selected.
    staff_id = models.PositiveIntegerField()
    flags = models.PositiveIntegerField()

    class Meta:
        managed = False
        db_table = "ost_team_member"
        unique_together = (("team_id", "staff_id"),)


class OstThread(models.Model):
    object_id = models.PositiveIntegerField()
    object_type = models.CharField(max_length=1)
    extra = models.TextField(blank=True, null=True)
    lastresponse = models.DateTimeField(blank=True, null=True)
    lastmessage = models.DateTimeField(blank=True, null=True)
    created = models.DateTimeField()

    @property
    def ticket(self):
        if self.object_type == OstObjectType.TICKET:
            return OstTicket.objects.get(pk=self.object_id)
        return None

    @property
    def task(self):
        if self.object_type == OstObjectType.TASK:
            return OstTask.objects.get(pk=self.object_id)
        return None

    @property
    def creator(self):
        entry = self.ostthreadentry_set.first()
        return entry.staff or entry.user

    class Meta:
        managed = False
        db_table = "ost_thread"


class OstThreadCollaborator(models.Model):
    flags = models.PositiveIntegerField()
    # thread_id = models.PositiveIntegerField()
    thread = models.ForeignKey(OstThread, on_delete=models.DO_NOTHING)
    # user_id = models.PositiveIntegerField()
    user = models.ForeignKey("OstUser", on_delete=models.DO_NOTHING)
    role = models.CharField(max_length=1)
    created = models.DateTimeField()
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_thread_collaborator"
        unique_together = (("thread_id", "user_id"),)


class OstThreadEntry(models.Model):
    pid = models.PositiveIntegerField()
    # thread_id = models.PositiveIntegerField()
    thread = MySQLZeroForeignKey(OstThread, on_delete=models.DO_NOTHING)
    # staff_id = models.PositiveIntegerField()
    staff = MySQLZeroForeignKey(OstStaff, on_delete=models.DO_NOTHING)
    # user_id = models.PositiveIntegerField()
    user = MySQLZeroForeignKey("OstUser", on_delete=models.DO_NOTHING)
    type = models.CharField(max_length=1)
    flags = models.PositiveIntegerField()
    poster = models.CharField(max_length=128)
    editor = models.PositiveIntegerField(blank=True, null=True)
    editor_type = models.CharField(max_length=1, blank=True, null=True)
    source = models.CharField(max_length=32)
    title = models.CharField(max_length=255, blank=True, null=True)
    body = models.TextField()
    format = models.CharField(max_length=16)
    ip_address = models.CharField(max_length=64)
    # Some versions of OsTicket don't have this field
    # extra = models.TextField(blank=True, null=True)
    recipients = models.TextField(blank=True, null=True)
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    class Meta:
        managed = False
        db_table = "ost_thread_entry"

    def poster_name(self) -> Optional[str]:
        if self.poster:
            if not hasattr(self, "nemo_user"):
                self.nemo_user = User.objects.filter(username=self.poster).values("first_name", "last_name").first()
            if self.nemo_user:
                return f"{self.nemo_user.get('first_name')} {self.nemo_user.get('last_name')}"
            else:
                return self.poster
        return None


class OstThreadEntryEmail(models.Model):
    # thread_entry_id = models.PositiveIntegerField()
    thread_entry = models.ForeignKey(OstThreadEntry, on_delete=models.DO_NOTHING)
    mid = models.CharField(max_length=255)
    headers = models.TextField(blank=True, null=True)

    class Meta:
        managed = False
        db_table = "ost_thread_entry_email"


class OstThreadEntryMerge(models.Model):
    # thread_entry_id = models.PositiveIntegerField()
    thread_entry = models.ForeignKey(OstThreadEntry, on_delete=models.DO_NOTHING)
    data = models.TextField(blank=True, null=True)

    class Meta:
        managed = False
        db_table = "ost_thread_entry_merge"


class OstThreadEvent(models.Model):
    # thread_id = models.PositiveIntegerField()
    thread = models.ForeignKey(OstThread, on_delete=models.DO_NOTHING)
    thread_type = models.CharField(max_length=1)
    # event_id = models.PositiveIntegerField(blank=True, null=True)
    event = models.ForeignKey("OstThreadEvent", blank=True, null=True, on_delete=models.DO_NOTHING)
    staff_id = models.PositiveIntegerField()
    team_id = models.PositiveIntegerField()
    dept_id = models.PositiveIntegerField()
    topic_id = models.PositiveIntegerField()
    data = models.CharField(max_length=1024, blank=True, null=True, db_comment="Encoded differences")
    username = models.CharField(max_length=128)
    uid = models.PositiveIntegerField(blank=True, null=True)
    uid_type = models.CharField(max_length=1)
    annulled = models.PositiveIntegerField()
    timestamp = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_thread_event"


class OstThreadReferral(models.Model):
    thread_id = models.PositiveIntegerField()
    object_id = models.PositiveIntegerField()
    object_type = models.CharField(max_length=1)
    created = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_thread_referral"
        unique_together = (("object_id", "object_type", "thread_id"),)


class OstTicket(models.Model):
    ticket_id = models.AutoField(primary_key=True)
    ticket_pid = models.PositiveIntegerField(blank=True, null=True)
    number = models.CharField(max_length=20, blank=True, null=True)
    # user_id = models.PositiveIntegerField()
    user = models.ForeignKey("OstUser", on_delete=models.DO_NOTHING)
    # user_email_id = models.PositiveIntegerField()
    user_email = MySQLZeroForeignKey("OstUserEmail", on_delete=models.DO_NOTHING)
    # status_id = models.PositiveIntegerField()
    status = models.ForeignKey("OstTicketStatus", on_delete=models.DO_NOTHING)
    # dept_id = models.PositiveIntegerField()
    dept = models.ForeignKey(OstDepartment, null=True, blank=True, on_delete=models.DO_NOTHING)
    # sla_id = models.PositiveIntegerField()
    sla = models.ForeignKey(OstSla, on_delete=models.DO_NOTHING)
    # topic_id = models.PositiveIntegerField()
    topic = models.ForeignKey(OstHelpTopic, null=True, blank=True, on_delete=models.DO_NOTHING)
    # staff_id = models.PositiveIntegerField()
    staff = MySQLZeroForeignKey(OstStaff, on_delete=models.DO_NOTHING)
    # team_id = models.PositiveIntegerField()
    team = MySQLZeroForeignKey(OstTeam, on_delete=models.DO_NOTHING)
    # email_id = models.PositiveIntegerField()
    email = MySQLZeroForeignKey(OstEmail, on_delete=models.DO_NOTHING)
    # lock_id = models.PositiveIntegerField()
    lock = MySQLZeroOneToOne(OstLock, on_delete=models.DO_NOTHING)
    flags = models.PositiveIntegerField()
    sort = models.PositiveIntegerField()
    ip_address = models.CharField(max_length=64)
    source = models.CharField(max_length=5)
    source_extra = models.CharField(max_length=40, blank=True, null=True)
    isoverdue = models.PositiveIntegerField()
    isanswered = models.PositiveIntegerField()
    duedate = models.DateTimeField(blank=True, null=True)
    est_duedate = models.DateTimeField(blank=True, null=True)
    reopened = models.DateTimeField(blank=True, null=True)
    closed = models.DateTimeField(blank=True, null=True)
    lastupdate = models.DateTimeField(blank=True, null=True)
    created = models.DateTimeField()
    updated = models.DateTimeField()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.form_entry_values = {}

    @property
    def osttask_set(self):
        return OstTask.objects.filter(object_id=self.ticket_id, object_type=OstObjectType.TICKET)

    def get_osticket_absolute_url(self):
        from NEMO_osticket.views.osticket import get_os_ticket_service

        osticket_service = get_os_ticket_service()
        url = osticket_service.get("url")
        if osticket_service.get("available", False) and url:
            return url + "/scp/tickets.php?id={}".format(self.ticket_id)
        return None

    def creator_name(self) -> Optional[str]:
        if self.user:
            if not hasattr(self, "nemo_user"):
                self.nemo_user = User.objects.filter(username=self.user.name).values("first_name", "last_name").first()
            if self.nemo_user:
                return f"{self.nemo_user.get('first_name')} {self.nemo_user.get('last_name')}"
            else:
                return self.user.name
        return None

    def subject(self) -> Optional[str]:
        if not hasattr(self, "ostthreadentry_set"):
            if not hasattr(self, "ostthread_set"):
                self.ostthread_set = set(
                    OstThread.objects.filter(
                        object_id__in=self.ticket_id, object_type=OstObjectType.TICKET
                    ).prefetch_related("ostthreadentry_set")
                )
            thread = next(iter(self.ostthread_set))
            self.ostthreadentry_set = thread.ostthreadentry_set.all()
        if self.ostthreadentry_set:
            entry = next(iter(self.ostthreadentry_set))
            if entry:
                return entry.title
        return None

    def assigned_to(self) -> Optional[str]:
        if self.staff:
            return f"{self.staff.firstname} {self.staff.lastname}"
        return None

    class Meta:
        managed = False
        db_table = "ost_ticket"


class OstTicketPriority(models.Model):
    priority_id = models.AutoField(primary_key=True)
    priority = models.CharField(unique=True, max_length=60)
    priority_desc = models.CharField(max_length=30)
    priority_color = models.CharField(max_length=7)
    priority_urgency = models.PositiveIntegerField()
    ispublic = models.IntegerField()

    class Meta:
        managed = False
        db_table = "ost_ticket_priority"


class OstTicketStatus(models.Model):
    name = models.CharField(unique=True, max_length=60)
    state = models.CharField(max_length=16, blank=True, null=True)
    mode = models.PositiveIntegerField()
    flags = models.PositiveIntegerField()
    sort = models.PositiveIntegerField()
    properties = models.TextField()
    created = models.DateTimeField()
    updated = models.DateTimeField()

    class Meta:
        managed = False
        db_table = "ost_ticket_status"


class OstTranslation(models.Model):
    object_hash = models.CharField(max_length=16, db_collation="ascii_general_ci", blank=True, null=True)
    type = models.CharField(max_length=8, blank=True, null=True)
    flags = models.PositiveIntegerField()
    revision = models.PositiveIntegerField(blank=True, null=True)
    agent_id = models.PositiveIntegerField()
    lang = models.CharField(max_length=16)
    text = models.TextField()
    source_text = models.TextField(blank=True, null=True)
    updated = models.DateTimeField(blank=True, null=True)

    class Meta:
        managed = False
        db_table = "ost_translation"


class OstUser(models.Model):
    # org_id = models.PositiveIntegerField()
    org = MySQLZeroForeignKey(OstOrganization, on_delete=models.DO_NOTHING)
    default_email = MySQLZeroForeignKey("OstUserEmail", on_delete=models.DO_NOTHING)
    status = models.PositiveIntegerField()
    name = models.CharField(max_length=128)
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    class Meta:
        managed = False
        db_table = "ost_user"


class OstUserAccount(models.Model):
    # user_id = models.PositiveIntegerField()
    user = models.OneToOneField(OstUser, null=True, blank=True, on_delete=models.DO_NOTHING)
    status = models.PositiveIntegerField()
    timezone = models.CharField(max_length=64, blank=True, null=True)
    lang = models.CharField(max_length=16, blank=True, null=True)
    username = models.CharField(unique=True, max_length=64, blank=True, null=True)
    passwd = models.CharField(max_length=128, db_collation="ascii_bin", blank=True, null=True)
    backend = models.CharField(max_length=32, blank=True, null=True)
    extra = models.TextField(blank=True, null=True)
    registered = models.DateTimeField(blank=True, null=True)

    class Meta:
        managed = False
        db_table = "ost_user_account"


class OstUserEmail(models.Model):
    # user_id = models.PositiveIntegerField()
    user = models.ForeignKey(OstUser, on_delete=models.DO_NOTHING)
    flags = models.PositiveIntegerField()
    address = models.CharField(unique=True, max_length=255)

    class Meta:
        managed = False
        db_table = "ost_user_email"



from NEMO.models import Tool
from django.apps import apps
from django.test import TestCase

from NEMO_osticket.customizations import OsTicketCustomization


class OsTicketTest(TestCase):

    def test_plugin_is_installed(self):
        assert apps.is_installed("NEMO_osticket")

    def test_matching(self):
        tool = Tool(id=2, name="Test tool", _serial="123")
        # no matching regular expression -> no match
        self.assertFalse(OsTicketCustomization.is_tool_match(tool, "2"))
        # now let's set one
        OsTicketCustomization.set("osticket_tool_matching_property_extract_re", "(\d+)")
        self.assertTrue(
            OsTicketCustomization.is_tool_match(
                tool, "2", OsTicketCustomization.get("osticket_tool_matching_property_extract_re")
            )
        )
        # try with serial number for example
        OsTicketCustomization.set("osticket_tool_matching_nemo_property_template", "{{ tool.serial }}")
        OsTicketCustomization.set("osticket_tool_matching_property_extract_re", "(\d+):")
        self.assertTrue(
            OsTicketCustomization.is_tool_match(
                tool, "123: This tool", OsTicketCustomization.get("osticket_tool_matching_property_extract_re")
            )
        )



from django.urls import include, path

from NEMO_osticket.views import osticket

urlpatterns = [
    path(
        "osticket/",
        include(
            [
                path("tickets/", osticket.tickets, name="osticket_tickets"),
                path("ticket/<int:ticket_id>/", osticket.ticket_details, name="osticket_ticket_details"),
                path("ticket/<int:ticket_id>/reply/", osticket.reply_to_ticket, name="osticket_reply_to_ticket"),
                path("create_ticket/<int:tool_id>/", osticket.create_tool_ticket, name="osticket_create_tool_ticket"),
                path("create_ticket/", osticket.create_ticket, name="osticket_create_ticket"),
                path(
                    "tool_status_tickets/<int:tool_id>/",
                    osticket.tool_status_tickets,
                    name="osticket_tool_status_tickets",
                ),
            ]
        ),
    ),
    # Override tool status to add our own osticket tab
    path("tool_status/<int:tool_id>/", osticket.tool_status, name="tool_status"),
    # Override searching in past comments and tasks
    path("past_comments_and_tasks/", osticket.past_comments_and_tasks, name="past_comments_and_tasks"),
    path(
        "ten_most_recent_past_comments_and_tasks/<int:tool_id>/",
        osticket.ten_most_recent_past_comments_and_tasks,
        name="ten_most_recent_past_comments_and_tasks",
    ),
]



import re
from datetime import timedelta
from itertools import chain
from logging import getLogger
from typing import Iterable

from NEMO.forms import nice_errors
from NEMO.models import Comment, Task, Tool, UsageEvent, User
from NEMO.utilities import (
    export_format_datetime,
    extract_optional_beginning_and_end_times,
    format_datetime,
    render_combine_responses,
)
from NEMO.views.pagination import SortedPaginator
from NEMO.views.tool_control import tool_status as original_tool_status
from django.conf import settings
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse, HttpResponseBadRequest, HttpResponseForbidden
from django.shortcuts import get_object_or_404, redirect, render
from django.utils import timezone
from django.utils.html import strip_tags
from django.views.decorators.http import require_GET, require_POST, require_http_methods

from NEMO_osticket.customizations import OsTicketCustomization
from NEMO_osticket.exceptions import OsTicketException
from NEMO_osticket.forms import OsTicketForm, OsTicketSearchForm
from NEMO_osticket.models import OstThreadEntryType, OstTicket
from NEMO_osticket.osticket_api import (
    OsTicketSearchObject,
    add_reply_to_ticket,
    create_os_ticket,
    get_os_ticket_service,
    get_preloaded_os_tickets,
    get_ticket_form_field_labels,
)

osticket_logger = getLogger(__name__)


@login_required
@require_GET
def tickets(request):
    user: User = request.user
    search_form = OsTicketSearchForm(request.GET or None, initial={"email": user.email, "is_open": True})
    ticket_field_labels = get_ticket_form_field_labels()
    dictionary = {
        "form": search_form,
        "search_done": request.GET,
        "page": [],
        "ticket_field_labels": ticket_field_labels,
    }

    if search_form.is_valid():
        matching_tickets = search_form.get_search_object().search_tickets(preload=False)
        ticket_list = SortedPaginator(matching_tickets, request, order_by="-created").get_current_page()
        dictionary["page"] = ticket_list
        dictionary["tickets"] = get_preloaded_os_tickets(
            [ticket.ticket_id for ticket in ticket_list], order_by=ticket_list.paginator.order_by
        )
    return render(request, "NEMO_osticket/tickets/ticket_search.html", dictionary)


@login_required
@require_GET
def ticket_details(request, ticket_id):
    get_object_or_404(OstTicket, pk=ticket_id)
    loaded_tickets = get_preloaded_os_tickets([ticket_id])
    ticket_field_labels = get_ticket_form_field_labels()
    dictionary = {
        "ticket": next(iter(loaded_tickets)) if loaded_tickets else None,
        "ticket_field_labels": ticket_field_labels,
    }
    return render(request, "NEMO_osticket/ticket_details/ticket_details.html", dictionary)


@login_required
@require_GET
def tool_status(request, tool_id):
    original_response = original_tool_status(request, tool_id)
    no_osticket = not get_os_ticket_service().get("available", False) and not settings.DATABASES.get("osticket", False)
    if no_osticket or original_response.status_code != 200:
        return original_response

    dictionary = {
        "tool_id": tool_id,
        "replace_problems_tab": OsTicketCustomization.get_bool("osticket_tool_control_replace_problem_tab"),
        "tab_name": OsTicketCustomization.get("osticket_tool_control_tab_name"),
    }

    return render_combine_responses(
        request, original_response, "NEMO_osticket/tool_control/osticket_tab.html", dictionary
    )


@login_required
@require_GET
def tool_status_tickets(request, tool_id):
    tool = get_object_or_404(Tool, pk=tool_id)
    create_ticket_include_usage_events = OsTicketCustomization.get_bool("osticket_create_ticket_include_usage_events")
    past_week_usage_events = UsageEvent.objects.none()
    if create_ticket_include_usage_events:
        # since a week ago
        start = timezone.now() - timedelta(days=7)
        past_week_usage_events = UsageEvent.objects.filter(
            tool_id=tool_id, start__gte=start, end__isnull=False
        ).order_by("-start")
        past_week_usage_events = past_week_usage_events.prefetch_related("user")

    dictionary = {
        "usage_events": past_week_usage_events,
        "osticket_api_available": get_os_ticket_service().get("available", False),
        "tickets": OsTicketSearchObject(tool=tool, is_open=True).search_tickets(),
        "form": OsTicketForm(),
        "tool_id": tool_id,
        "ticket_field_labels": get_ticket_form_field_labels(),
    }
    return render(request, "NEMO_osticket/tool_control/osticket_tab_content.html", dictionary)


@login_required
@require_http_methods(["GET", "POST"])
def create_ticket(request):
    form = OsTicketForm(request.POST or None, include_usage_events=False)
    dictionary = {"form": form}
    if form.is_valid():
        try:
            create_os_ticket(request, form.cleaned_data, None)
            messages.success(request, "Your ticket was successfully created")
            return redirect("osticket_tickets")
        except OsTicketException as e:
            osticket_logger.exception(e)
            form.add_error(None, f"There was an error creating the ticket: {str(e)}")
    return render(request, "NEMO_osticket/tickets/create_ticket.html", dictionary)


@login_required
@require_POST
def create_tool_ticket(request, tool_id):
    tool = get_object_or_404(Tool, pk=tool_id)

    form = OsTicketForm(request.POST)
    if not form.is_valid():
        return HttpResponseBadRequest(nice_errors(form).as_ul())

    try:
        create_os_ticket(request, form.cleaned_data, tool)
        messages.success(request, "Your ticket was successfully created")
    except OsTicketException as e:
        osticket_logger.exception(e)
        messages.error(request, f"There was an error creating the ticket: {str(e)}")

    return redirect("tool_control")


@login_required
@require_POST
def reply_to_ticket(request, ticket_id):
    ticket = get_object_or_404(OstTicket, pk=ticket_id)
    user: User = request.user
    if not user.is_active:
        return HttpResponseForbidden("You are not authorized to reply to this ticket")
    message = request.POST.get("message")
    # Only staff can post internal notes
    internal = request.POST.get("internal", False) == "on" and user.is_any_part_of_staff
    if message:
        add_reply_to_ticket(request, ticket, user, message, internal)
    return redirect("osticket_ticket_details", ticket_id=ticket_id)


@login_required
@require_GET
def past_comments_and_tasks(request):
    user: User = request.user
    start, end = extract_optional_beginning_and_end_times(request.GET)
    search = request.GET.get("search")
    if not start and not end and not search:
        return HttpResponseBadRequest("Please enter a search keyword, start date or end date.")
    tool = get_object_or_404(Tool, pk=request.GET["tool_id"])
    search_object = OsTicketSearchObject(**{"tool": tool, "search": search, "start": start, "end": end})
    try:
        matching_tickets = search_object.search_tickets()
        tasks = Task.objects.filter(tool_id=tool.id)
        comments = Comment.objects.filter(tool_id=tool.id)
        if not user.is_staff:
            comments = comments.filter(staff_only=False)
        if start:
            tasks = tasks.filter(creation_time__gt=start)
            comments = comments.filter(creation_date__gt=start)
        if end:
            tasks = tasks.filter(creation_time__lt=end)
            comments = comments.filter(creation_date__lt=end)
        if search:
            tasks = tasks.filter(problem_description__icontains=search)
            comments = comments.filter(content__icontains=search)
    except:
        osticket_logger.exception("Task and comment lookup failed.")
        return HttpResponseBadRequest("Task and comment lookup failed.")
    past = list(chain(matching_tickets, tasks, comments))
    past.sort(
        key=lambda x: getattr(x, "created", None)
        or getattr(x, "creation_time", None)
        or getattr(x, "creation_date", None)
    )
    past.reverse()
    if request.GET.get("export"):
        return export_comments_and_tasks_to_text(request, past)
    dictionary = {"past": past}
    return render(request, "NEMO_osticket/tool_control/osticket_past_tasks_and_comments.html", dictionary)


@login_required
@require_GET
def ten_most_recent_past_comments_and_tasks(request, tool_id):
    user: User = request.user
    tool = get_object_or_404(Tool, pk=tool_id)
    matching_tickets = list(OsTicketSearchObject(tool=tool, is_open=True).search_tickets())[:10]
    tasks = Task.objects.filter(tool_id=tool_id).order_by("-creation_time")[:10]
    comments = Comment.objects.filter(tool_id=tool_id).order_by("-creation_date")
    if not user.is_staff:
        comments = comments.filter(staff_only=False)
    comments = comments[:10]
    past = list(chain(matching_tickets, tasks, comments))
    past.sort(
        key=lambda x: getattr(x, "created", None)
        or getattr(x, "creation_time", None)
        or getattr(x, "creation_date", None)
    )
    past.reverse()
    past = past[0:10]
    if request.GET.get("export"):
        return export_comments_and_tasks_to_text(request, past)
    dictionary = {"past": past}
    return render(request, "NEMO_osticket/tool_control/osticket_past_tasks_and_comments.html", dictionary)


def export_comments_and_tasks_to_text(request, comments_and_tasks: Iterable):
    topic_label = OsTicketCustomization.get("osticket_create_ticket_help_topic_label")
    content = "No tickets, tasks or comments were created between these dates." if not comments_and_tasks else ""
    for item in comments_and_tasks:
        if isinstance(item, OstTicket):
            ticket: OstTicket = item
            if ticket.topic_id:
                content += f"{topic_label}: {ticket.topic.topic}\n"
            subject_label_and_value = ticket.form_entry_values.get("subject")
            if subject_label_and_value:
                subject_label, subject_value = subject_label_and_value
                content += f"{subject_label}: {subject_value}\n"
            for thread_entry in ticket.ostthreadentry_set:
                if thread_entry.type != OstThreadEntryType.INTERNAL_NOTE or request.user.is_staff:
                    content += f"\nOn {format_datetime(ticket.created)}\n"
                    content += f"{html_to_text(thread_entry.body)}\n"
        if isinstance(item, Comment):
            comment: Comment = item
            staff_only = "staff only " if comment.staff_only else ""
            content += f"On {format_datetime(comment.creation_date)} {comment.author} wrote this {staff_only}comment:\n"
            content += f"{comment.content}\n"
            if comment.hide_date:
                content += f"{comment.hidden_by} hid the comment on {format_datetime(comment.hide_date)}.\n"
        elif isinstance(item, Task):
            task: Task = item
            content += f"On {format_datetime(task.creation_time)} {task.creator} created this task:\n"
            if task.problem_category:
                content += f"{task.problem_category.name}\n"
            if task.force_shutdown:
                content += "\nThe tool was shut down because of this task.\n"
            if task.progress_description:
                content += f"\n{task.progress_description}\n"
            if task.resolved:
                resolution_category = f"({task.resolution_category}) " if task.resolution_category else ""
                content += (
                    f"\nResolved {resolution_category}On {format_datetime(task.resolution_time)} by {task.resolver }.\n"
                )
                if task.resolution_description:
                    content += f"{task.resolution_description}\n"
            elif task.cancelled:
                content += f"\nCancelled On {format_datetime(task.resolution_time)} by {task.resolver}.\n"
        content += "\n---------------------------------------------------\n\n"
    response = HttpResponse(content, content_type="text/plain")
    response["Content-Disposition"] = "attachment; filename={0}".format(
        f"comments_and_tasks_export_{export_format_datetime()}.txt"
    )
    return response


def html_to_text(html):
    # Replace <br> or <br/> with newline
    text = re.sub(r"<br\s*/?>", "\n", html, flags=re.IGNORECASE)
    # Remove all other HTML tags
    text = strip_tags(text)
    return text


# TODO: add a hook to kiosk as well
